// main.c - UF GBA slide show
// Copyright (c) 2001 by Ultra Force Development
//
// This program may be distrubuted freely.
//
// version 1
// - initial version
//
// for use with gcc 3.0
//

#include "gba.h"

// initial delay for key check
#define KEYWAITDELAY 32
// delay when key is kept pressed down
#define KEYREPEATDELAY 15

/**
 * Debug print, for use with Mappy
 */
volatile void dprint(char *textline) {
   asm volatile("
    mov r2, %0
    ldr r0, =0xC0DED00D
    mov r1, #0
    and r0, r0, r0
    " 
   : 
   : "r" (textline) 
   : "r0", "r1", "r2"
   );
}

/**
 * A vertical blank ?
 *
 * @returns 0: no blank, <>0: a blank
 */
u16 isVblank()
{
  return( DISPSTAT & STAT_VBL_MASK );
}

/**
 * Wait for the next vertical blank
 */
void waitVSync()
{
  while(isVblank());
  while(!isVblank());
}

/**
 * Copy a memory block, use a minumum of 16 bit transfers to copy data.
 *
 * @param pTarget - target address
 * @param pSource - source address
 * @param uCount - number of bytes (!) to copy, this should be an even number.
 */
void memwcpy(void *pTarget, void *pSource, u32 uCount)
{
  uCount = uCount / 2;
  u16 *pt = (u16 *) pTarget;
  u16 *ps = (u16 *) pSource;
  for(; uCount > 0; uCount--) *pt++ = *ps++;
}

/**
 * Copy an image to the screen.
 */
void copyImage(u16 *pTarget, u16 *pSource)
{
  memwcpy(pTarget, pSource, 240*160*2);
}

/**
 * Format of slideshow data:
 * u32 count, number of images
 * u32 start[], start of every images (count entries)
 * u16 pixels, data for every image
 */
typedef struct tagImageList
{
  u32 count;
  u16 *start[];
} ImageList;

/**
 * ImageInfo is used by the slideshow builder to fill in the start location of the image.
 */
typedef struct tagImageInfo
{
  u32 magic1;
  u32 magic2;
  ImageList *list;
} ImageInfo;

// will be update by slideshow with start of ImageList structure.
ImageInfo imageinfo = {0x42474655, 0x535341, 0};

/**
 * Main, the start of the application.
 */
int AgbMain()
{
  // pScreen points to the screen
  u16 *pScreen;
  int imgCount;
  int imgCurrent;
  int imgNew;
  u32 keyDelay = KEYWAITDELAY;
  u32 keyDelayCount;

  // switch to display mode 3, enable bg2. Also disable the screen, until it has been initialized.
  // sprites are disabled, use DISP_OBJ to enable them
  DISPCNT = DISP_MODE(3) | DISP_BG2 | DISP_OAMSEQ | DISP_DISABLE;

  // just to be sure
  BG2CNT = BG_CHARBASE(0x0) | BG_MAPBASE(0x00) | BG_COLOUR_256x1 | BG_T256x512 | BG_PRIORITY(1);  

  // pointers screen
  pScreen = (u16*) VRAM_BASE;

  // all backgrounds are positioned at 0,0
  BG2X = 0;
  BG2Y = 0;

  // don't do anything if there are no images
  while (imageinfo.list == 0);

  // retrieve count
  imgCount = imageinfo.list->count;

  // start with image 0
  imgCurrent = 0;

  // copy image
  copyImage(pScreen, imageinfo.list->start[imgCurrent]);

  // enable screen
  DISPCNT = DISPCNT & ~DISP_DISABLE;

  // react to joypad presses
  while(1)
  {
    // copy current index
    imgNew = imgCurrent;
    if (!(PADREG & PAD_RIGHT)) imgNew++;
    if (!(PADREG & PAD_LEFT)) imgNew--;
    // wrap around if neccesary
    if (imgNew < 0) imgNew = imgCount - 1;
    if (imgNew >= imgCount) imgNew = 0;
    // copy image
    if (imgNew != imgCurrent)
    {
      waitVSync();
      copyImage(pScreen, imageinfo.list->start[imgNew]);
      imgCurrent = imgNew;
      // wait keyDelay frames
      for(keyDelayCount = keyDelay; keyDelayCount > 0; keyDelayCount--) 
      {
        // next time use shorter delay
        keyDelay = KEYREPEATDELAY;
        waitVSync();
        // both keys are released ?
        if ((PADREG & PAD_RIGHT) && (PADREG & PAD_LEFT)) 
        {
          // use initial delay 
          keyDelay = KEYWAITDELAY;
          // break from loop
          break;
        }
      }
    }
  }
}
